from ingenialink.canopen.register import CanopenRegister
from ingenialink.enums.register import (
    REG_ACCESS,
    REG_ADDRESS_TYPE,
    REG_DTYPE,
    REG_PHY,
    RegCyclicType,
)


class EthercatRegister(CanopenRegister):
    pass
